#!/usr/bin/env python
import os
import sys
import time
import glob
import shutil
import urlparse
import urllib2
import pwd

from charmhelpers.core.hookenv import (
    log,
    Hooks,
    config,
)

from charmhelpers.core.host import (
    lsb_release,
    rsync,
    mkdir,
)

from charmhelpers.fetch import (
    add_source,
    install_remote,
)

hooks = Hooks()
hook_dir = os.path.abspath(os.path.dirname(__file__))
charm_dir = os.path.dirname(hook_dir)
template_dir = os.path.join(charm_dir, 'templates')
fetch_dir = os.path.join(charm_dir, 'fetched')
revno_file = '/var/lib/juju/content-revno'


@hooks.hook('upgrade-charm')
@hooks.hook('install')
def install():
    log("No action needed")


def prune_dirs(dir_glob, keep_dirs):
    direcs = []
    for direc in glob.glob(dir_glob):
        direcs.append(direc)
    number_to_delete = len(direcs) - keep_dirs
    direcs.sort()
    direcs.reverse()
    delete_count = 0
    for direc in direcs:
        if delete_count >= number_to_delete:
            shutil.rmtree(direc)
        delete_count += 1


def get_current_revno():
    with open(revno_file, 'r') as f:
        revno = f.read()
    if revno == '':
        return int(-1)
    else:
        return int(revno)


def set_revno(revno):
    with open(revno_file, 'w') as content_file:
        content_file.write(str(revno))


def setup_ssh(src_server):
    euser = pwd.getpwuid(os.geteuid())[0]
    euser_home = os.path.expanduser("~" + euser)
    ssh_dir = os.path.join(euser_home, '.ssh')
    ssh_config = os.path.join(ssh_dir, 'config')
    ssh_config_tmpl = os.path.join(template_dir, 'ssh_config.tmpl')
    fetcher_host_key_file = os.path.join(ssh_dir, 'fetcher_host_key')
    charm_key_dir = os.path.join(charm_dir, 'ssh-key')
    charm_key = os.path.join(charm_key_dir, 'content_fetcher_rsa')
    installed_key = os.path.join(ssh_dir, 'content_fetcher_rsa')
    remote_host_key = config().get('ssh_host_key')
    # Create .ssh in homedir if needed
    if not os.path.exists(ssh_dir):
        os.mkdir(ssh_dir, mode=0700)
    # Copy in the private key from the charm
    if os.path.isfile(charm_key):
        with open(charm_key, 'r') as content_file:
            private_key = content_file.read()
        with open(installed_key, 'w') as content_file:
            os.chmod(installed_key, 0600)
            content_file.write(private_key)
    # If a hostkey has been provided write it out to its own file to be
    # referenced by the users ssh config later
    if len(remote_host_key) > 0:
        with open(fetcher_host_key_file, 'w') as f:
            os.chmod(fetcher_host_key_file, 0644)
            f.write(remote_host_key + '\n')
    with open(ssh_config_tmpl, 'r') as f:
        template = f.read()
    template_args = {'src_server': src_server, 'key_file': installed_key}
    open_mode = 'w'
    if os.path.isfile(ssh_config):
        for line in open(ssh_config):
            if "Host " + src_server in line:
                return
        open_mode = 'a'
    # If a hostkey has been provided check against the file containing it
    # otherwise don't verify hostkeys
    with open(ssh_config, open_mode) as f:
        os.chmod(ssh_config, 0644)
        f.write(template.format(**template_args))
        if len(remote_host_key) > 0:
            f.write('   UserKnownHostsFile ' + fetcher_host_key_file + '\n')
        else:
            f.write('   StrictHostKeyChecking no')


@hooks.hook('config-changed')
def config_changed():
    trigger_revno = config().get('deploy_trigger')
    if os.path.isfile(revno_file):
        current_revno = get_current_revno()
        if trigger_revno > current_revno:
            deploy_archive()
        else:
            log("The deploy_trigger matches previous deploy.")
            log("Not updating content.")
    else:
        deploy_archive()


def deploy_archive():
    trigger_revno = config().get('deploy_trigger')
    arch_url = config().get('archive_location')
    if len(arch_url) == 0:
        print "No archive location specified"
        return
    url_parts = urlparse.urlparse(arch_url)
    src_type = url_parts[0]
    dst_dir = config().get('dest_dir')
    if len(src_type) == 0 and len(config().get('tenant_name')) > 0:
        # XXX ok, this isn't nice. But we need the swiftclient for precise and
        # the cloud archive is a safe place to get it
        if lsb_release()['DISTRIB_CODENAME'] == 'precise':
            add_source('cloud:precise-updates/folsom', '5EDB1B62EC4926EA')
        import swift
        sc = swift.SwiftFetchHandler(config().get('auth_url'),
                                     config().get('username'),
                                     config().get('secret'),
                                     config().get('tenant_name'))
        archive_dir = sc.install(config().get('container_name'),
                                 arch_url)
    elif src_type == 'bzr+ssh':
        # Clean out the tmp area that may hold an old branch
        for direc in glob.glob(os.path.join(fetch_dir, '*')):
            if os.path.isdir(direc):
                shutil.rmtree(direc)
        url_parts = urlparse.urlparse(arch_url)
        auth, src_server = urllib2.splituser(url_parts.netloc)
        setup_ssh(src_server)
        archive_dir = install_remote(arch_url)
    else:
        archive_dir = install_remote(arch_url)
    if os.path.exists(dst_dir):
        backup_postfix = time.strftime("-%Y%m%d-%H%M%S", time.gmtime())
        backup_dir = dst_dir.rstrip('/') + backup_postfix
        log("Backing up old content to " + backup_dir)
        mkdir(backup_dir, perms=0755)
        rsync(dst_dir + '/', backup_dir)
    else:
        mkdir(dst_dir, perms=0755)
    log("Installing new content")
    rsync(archive_dir + '/', dst_dir)
    backup_dir_glob = dst_dir.rstrip('/') + "-*"
    prune_dirs(backup_dir_glob, config().get('keep_dir_count'))
    set_revno(trigger_revno)


if __name__ == "__main__":
    hooks.execute(sys.argv)
