#!/usr/bin/python
import os
import tempfile
import argparse
import socket
import shutil
import tarfile
import time
import sys
import pwd
import string
import random

from os.path import dirname
from zope.interface import implements
from twisted.internet import reactor, ssl
from twisted.web import server, resource, guard, static
from twisted.cred.portal import IRealm, Portal
from twisted.protocols.ftp import FTPFactory, FTPRealm
from twisted.cred.checkers import AllowAnonymousAccess

from twisted.cred.checkers import \
    InMemoryUsernamePasswordDatabaseDontUse as InMemoryDB
from Crypto.PublicKey import RSA


def get_listen_ip():
    return socket.gethostbyname(socket.gethostname())


def create_work_area():
    return tempfile.mkdtemp()


def create_test_html(pub_dir, access_type):
    test_file = os.path.join(pub_dir, 'index.html')
    f = open(test_file, 'w')
    f.write('<html><body><h1>Website collected from %s at %s</h1>\
             </body></html>\n' %
            (access_type, time.strftime("%Y/%m/%d %H:%M:%S", time.gmtime())))
    f.close()
    return test_file


def create_test_tarball(pub_dir, access_type):
    tar_file = os.path.join(pub_dir, 'content.tar')
    test_file = create_test_html(pub_dir, access_type)
    tar = tarfile.open(tar_file, "w")
    tar.add(test_file, arcname="index.html")
    tar.close()
    os.remove(test_file)


def create_test_bzrrepo(pub_dir, access_type):
    from bzrlib.bzrdir import BzrDir
    branch = BzrDir.create_branch_convenience(work_area, force_new_tree=True)
    test_file = create_test_html(work_area, 'bzr+ssh')
    branch.get_config().set_user_option('email', u'Branch Identity' +
                                        'content-fetcher@%s' % get_listen_ip())
    wtree = branch.bzrdir.open_workingtree()
    wtree.smart_add([test_file])
    wtree.commit('Commiting test data')


def tidyup(work_area):
    shutil.rmtree(work_area)


def get_charm_root():
    return dirname(dirname(os.path.realpath(__file__)))


def display_juju_set_cmd(src_uri=None):
    if src_uri:
        print "juju set content-fetcher archive_location=%s" % (src_uri)
    print "juju set content-fetcher deploy_trigger=$(juju get" + \
          " content-fetcher | grep -A6 deploy_trigger | " + \
          " awk '/value:/ {print $2 + 1}')"
    sys.stdout.flush()


class SimpleRealm(object):
    implements(IRealm)

    def requestAvatar(self, avatarId, mind, *interfaces):
        if resource.IResource in interfaces:
            return resource.IResource, root, lambda: None
        raise NotImplementedError()


def start_http_server(work_area, port, proto, **users):
    if users:
        create_test_tarball(work_area, '%s auth' % (proto))
        checkers = [InMemoryDB(**users)]
        for user in users:
            display_juju_set_cmd("%s://%s:%s@%s:%s/content.tar" %
                                 (proto, user, users[user],
                                  get_listen_ip(), port))
    else:
        create_test_tarball(work_area, '%s anon' % (proto))
        checkers = [AllowAnonymousAccess()]
        display_juju_set_cmd("%s://%s:%s/content.tar" %
                             (proto, get_listen_ip(), port))

    wrapper = guard.HTTPAuthSessionWrapper(
        Portal(SimpleRealm(), checkers),
        [guard.BasicCredentialFactory('test realm')])
    if proto == 'https':
        sslContext = ssl.DefaultOpenSSLContextFactory(
            os.path.join(get_charm_root(), 'testserver', 'website-key.pem'),
            os.path.join(get_charm_root(), 'testserver', 'website-cert.pem'),
        )
        reactor.listenSSL(port, server.Site(resource=wrapper),
                          contextFactory=sslContext)
    else:
        reactor.listenTCP(port, server.Site(resource=wrapper))
    try:
        reactor.run()
    finally:
        tidyup(work_area)


def start_ftp_server(work_area, port, **users):
    if users:
        for user in users:
            display_juju_set_cmd("ftp://%s:%s@%s:%s/content.tar" %
                                 (user, users[user], get_listen_ip(), port))
            os.mkdir(os.path.join(work_area, user))
            create_test_tarball(os.path.join(work_area, user), 'ftp auth')
    else:
        create_test_tarball(work_area, 'ftp anon')
        display_juju_set_cmd("ftp://%s:%s/content.tar" %
                             (get_listen_ip(), port))

    p = Portal(FTPRealm(work_area, userHome=work_area),
               (AllowAnonymousAccess(),
                InMemoryDB(**users),
                )
               )

    f = FTPFactory(p)

    try:
        reactor.listenTCP(port, f)
        reactor.run()
    finally:
        tidyup(work_area)


def setup_swift_test(work_area, swift_creds):
    import swiftclient
    create_test_tarball(work_area, 'swift')
    sc = swiftclient.client.Connection(authurl=swift_creds['auth_url'],
                                       user=swift_creds['username'],
                                       key=swift_creds['secret'],
                                       tenant_name=swift_creds['tenant_name'],
                                       auth_version='2.0')
    with open(os.path.join(work_area, 'content.tar'), 'r') as f:
        fcontent = f.read()
    sc.put_container(swift_creds['container_name'])
    sc.put_object(swift_creds['container_name'], 'content.tar', fcontent)
    juju_swift_params = ' '.join(["%s=%s" % (k, v)
                                  for k, v in swift_creds.iteritems()])
    print "juju set content-fetcher " + juju_swift_params
    display_juju_set_cmd()
    tidyup(work_area)


def setup_ssh_test(work_area, add_pub_key=False):
    create_test_bzrrepo(work_area, 'bzr+ssh')
    charm_key_dir = os.path.join(get_charm_root(), 'ssh-key')
    charm_key_file = os.path.join(charm_key_dir, 'content_fetcher_rsa')
    euser = pwd.getpwuid(os.geteuid())[0]
    euser_home = os.path.expanduser("~" + euser)
    ssh_dir = os.path.join(euser_home, '.ssh')
    local_privkey_file = os.path.join(ssh_dir, 'content_fetcher_rsa')
    local_authkey_file = os.path.join(ssh_dir, 'authorized_keys')
    print "To test this transport please:"
    if os.path.isfile(charm_key_file):
        with open(charm_key_file, 'r') as content_file:
            fcontent = content_file.read()
        key = RSA.importKey(fcontent)
    else:
        key = RSA.generate(2048)
        if os.path.isdir(charm_key_dir):
            with open(charm_key_file, 'w') as content_file:
                os.chmod(charm_key_file, 0600)
                content_file.write(key.exportKey('PEM') + '\n')
        else:
            with open(local_privkey_file, 'w') as content_file:
                os.chmod(local_privkey_file, 0600)
                content_file.write(key.exportKey('PEM') + '\n')

            print "0) Copy the key from %s to %s on the deployment server" % \
                (local_privkey_file, local_privkey_file)
    pubkey = key.publickey()

    print "* juju upgrade-charm content-fetcher"
    if add_pub_key:
        print local_authkey_file
        with open(local_authkey_file, 'a') as content_file:
            os.chmod(local_privkey_file, 0600)
            content_file.write(pubkey.exportKey('OpenSSH') + '\n')
    else:
        print "* Add public key below to authorized_keys file of the user"
        print "  used for the content fetch"
        print pubkey.exportKey('OpenSSH')
    display_juju_set_cmd('bzr+ssh://%s@%s/%s' % (os.environ['USER'],
                                                 get_listen_ip(),
                                                 work_area))


def setup_file_test(work_area):
    create_test_tarball(work_area, 'local')
    shutil.copy(work_area + '/content.tar', '/tmp/content.tar')
    print "juju scp /tmp/content.tar content-fetcher/0:/tmp/"
    display_juju_set_cmd('file:///tmp/content.tar')


parser = argparse.ArgumentParser()
parser.add_argument("listen_proto",
                    help="Proto type to test http, https, ftp, swift, bzr+ssh \
                            or file")
parser.add_argument("--port", type=int,
                    help="Port to listen on (http, https or ftp)")
parser.add_argument("--swift-user", help="swift username")
parser.add_argument("--swift-password", help="swift password")
parser.add_argument("--swift-authurl", help="swift authurl")
parser.add_argument("--swift-tenant", help="swift tenant")
parser.add_argument('--anon', action='store_true', default=False,
                    help="Setup anonymous access (http, https or ftp)")
parser.add_argument('--add-authkey', action='store_true', default=False,
                    help="Add generated public key to authorized_keys file")

args = parser.parse_args()

default_ports = {'ftp': 2121, 'http': 8080, 'https': 4343}
if args.port:
    listen_port = args.port
elif args.listen_proto in default_ports.keys():
    listen_port = default_ports[args.listen_proto]
if args.anon:
    creds = {}
else:
    password = ''.join(random.choice(string.ascii_letters +
                                     string.digits) for _ in range(8))
    creds = {'contentfetch': password}

work_area = create_work_area()
root = static.File(work_area)

if args.listen_proto == 'http':
    start_http_server(work_area, listen_port, args.listen_proto, **creds)
elif args.listen_proto == 'https':
    start_http_server(work_area, listen_port, args.listen_proto, **creds)
elif args.listen_proto == 'ftp':
    start_ftp_server(work_area, listen_port, **creds)
elif args.listen_proto == 'bzr+ssh':
    setup_ssh_test(work_area, args.add_authkey)
elif args.listen_proto == 'swift':
    if args.swift_user:
        swift_creds = {
            'username': args.swift_user,
            'secret': args.swift_password,
            'auth_url': args.swift_authurl,
            'tenant_name': args.swift_tenant,
        }
    elif 'OS_USERNAME' in os.environ:
        swift_creds = {
            'username': os.environ['OS_USERNAME'],
            'secret': os.environ['OS_PASSWORD'],
            'auth_url': os.environ['OS_AUTH_URL'],
            'tenant_name': os.environ['OS_TENANT_NAME'],
        }
    else:
        print "Could not discover swift creds"
        sys.exit(1)
    swift_creds['archive_location'] = "content.tar"
    swift_creds['container_name'] = "content-fetcher-test"
    setup_swift_test(work_area, swift_creds)
elif args.listen_proto == 'file':
    setup_file_test(work_area)
else:
    print args.listen_proto + " test server not currently supported"
sys.stdout.flush()
